@@ -1,11 +1,11 @@
+from abc import ABC, abstractmethod
 import base64
 import copy
-from abc import ABC, abstractmethod
 from enum import IntEnum
 from typing import (
     Any,
-    List,
     Dict,
+    List,
     Optional,
     Tuple,
     TypeVar,
@@ -14,6 +14,7 @@ from typing import (
 )
 
 from algosdk import abi, error, transaction
+from algosdk.transaction import GenericSignedTransaction
 from algosdk.abi.address_type import AddressType
 from algosdk.v2client import algod
 
@@ -23,27 +24,6 @@ ABI_RETURN_HASH = b"\x15\x1f\x7c\x75"
 T = TypeVar("T")
 
 
-class AtomicTransactionComposerStatus(IntEnum):
-    # BUILDING indicates that the atomic group is still under construction
-    BUILDING = 0
-
-    # BUILT indicates that the atomic group has been finalized,
-    # but not yet signed.
-    BUILT = 1
-
-    # SIGNED indicates that the atomic group has been finalized and signed,
-    # but not yet submitted to the network.
-    SIGNED = 2
-
-    # SUBMITTED indicates that the atomic group has been finalized, signed,
-    # and submitted to the network.
-    SUBMITTED = 3
-
-    # COMMITTED indicates that the atomic group has been finalized, signed,
-    # submitted, and successfully committed to a block.
-    COMMITTED = 4
-
-
 def populate_foreign_array(
     value_to_add: T, foreign_array: List[T], zero_value: Optional[T] = None
 ) -> int:
@@ -76,11 +56,234 @@ def populate_foreign_array(
     return offset + len(foreign_array) - 1
 
 
-GenericSignedTransaction = Union[
-    transaction.SignedTransaction,
-    transaction.LogicSigTransaction,
-    transaction.MultisigTransaction,
-]
+class AtomicTransactionComposerStatus(IntEnum):
+    # BUILDING indicates that the atomic group is still under construction
+    BUILDING = 0
+
+    # BUILT indicates that the atomic group has been finalized,
+    # but not yet signed.
+    BUILT = 1
+
+    # SIGNED indicates that the atomic group has been finalized and signed,
+    # but not yet submitted to the network.
+    SIGNED = 2
+
+    # SUBMITTED indicates that the atomic group has been finalized, signed,
+    # and submitted to the network.
+    SUBMITTED = 3
+
+    # COMMITTED indicates that the atomic group has been finalized, signed,
+    # submitted, and successfully committed to a block.
+    COMMITTED = 4
+
+
+class TransactionSigner(ABC):
+    """
+    Represents an object which can sign transactions from an atomic transaction group.
+    """
+
+    def __init__(self) -> None:
+        pass
+
+    @abstractmethod
+    def sign_transactions(
+        self, txn_group: List[transaction.Transaction], indexes: List[int]
+    ) -> List[GenericSignedTransaction]:
+        pass
+
+
+class AccountTransactionSigner(TransactionSigner):
+    """
+    Represents a Transaction Signer for an account that can sign transactions from an
+    atomic transaction group.
+
+    Args:
+        private_key (str): private key of signing account
+    """
+
+    def __init__(self, private_key: str) -> None:
+        super().__init__()
+        self.private_key = private_key
+
+    def sign_transactions(
+        self, txn_group: List[transaction.Transaction], indexes: List[int]
+    ) -> List[GenericSignedTransaction]:
+        """
+        Sign transactions in a transaction group given the indexes.
+
+        Returns an array of encoded signed transactions. The length of the
+        array will be the same as the length of indexesToSign, and each index i in the array
+        corresponds to the signed transaction from txnGroup[indexesToSign[i]].
+
+        Args:
+            txn_group (list[Transaction]): atomic group of transactions
+            indexes (list[int]): array of indexes in the atomic transaction group that should be signed
+        """
+        stxns = []
+        for i in indexes:
+            stxn = txn_group[i].sign(self.private_key)
+            stxns.append(stxn)
+        return stxns
+
+
+class LogicSigTransactionSigner(TransactionSigner):
+    """
+    Represents a Transaction Signer for a LogicSig that can sign transactions from an
+    atomic transaction group.
+
+    Args:
+        lsig (LogicSigAccount): LogicSig account
+    """
+
+    def __init__(self, lsig: transaction.LogicSigAccount) -> None:
+        super().__init__()
+        self.lsig = lsig
+
+    def sign_transactions(
+        self, txn_group: List[transaction.Transaction], indexes: List[int]
+    ) -> List[GenericSignedTransaction]:
+        """
+        Sign transactions in a transaction group given the indexes.
+
+        Returns an array of encoded signed transactions. The length of the
+        array will be the same as the length of indexesToSign, and each index i in the array
+        corresponds to the signed transaction from txnGroup[indexesToSign[i]].
+
+        Args:
+            txn_group (list[Transaction]): atomic group of transactions
+            indexes (list[int]): array of indexes in the atomic transaction group that should be signed
+        """
+        stxns: List[GenericSignedTransaction] = []
+        for i in indexes:
+            stxn = transaction.LogicSigTransaction(txn_group[i], self.lsig)
+            stxns.append(stxn)
+        return stxns
+
+
+class MultisigTransactionSigner(TransactionSigner):
+    """
+    Represents a Transaction Signer for a Multisig that can sign transactions from an
+    atomic transaction group.
+
+    Args:
+        msig (Multisig): Multisig account
+        sks (str): private keys of multisig
+    """
+
+    def __init__(self, msig: transaction.Multisig, sks: List[str]) -> None:
+        super().__init__()
+        self.msig = msig
+        self.sks = sks
+
+    def sign_transactions(
+        self, txn_group: List[transaction.Transaction], indexes: List[int]
+    ) -> List[GenericSignedTransaction]:
+        """
+        Sign transactions in a transaction group given the indexes.
+
+        Returns an array of encoded signed transactions. The length of the
+        array will be the same as the length of indexesToSign, and each index i in the array
+        corresponds to the signed transaction from txnGroup[indexesToSign[i]].
+
+        Args:
+            txn_group (list[Transaction]): atomic group of transactions
+            indexes (list[int]): array of indexes in the atomic transaction group that should be signed
+        """
+        stxns: List[GenericSignedTransaction] = []
+        for i in indexes:
+            mtxn = transaction.MultisigTransaction(txn_group[i], self.msig)
+            for sk in self.sks:
+                mtxn.sign(sk)
+            stxns.append(mtxn)
+        return stxns
+
+
+class EmptySigner(TransactionSigner):
+    def __init__(self) -> None:
+        super().__init__()
+
+    def sign_transactions(
+        self, txn_group: List[transaction.Transaction], indexes: List[int]
+    ) -> List[GenericSignedTransaction]:
+        stxns: List[GenericSignedTransaction] = []
+        for i in indexes:
+            stxns.append(transaction.SignedTransaction(txn_group[i], ""))
+        return stxns
+
+
+class TransactionWithSigner:
+    def __init__(
+        self, txn: transaction.Transaction, signer: TransactionSigner
+    ) -> None:
+        self.txn = txn
+        self.signer = signer
+
+
+class ABIResult:
+    def __init__(
+        self,
+        tx_id: str,
+        raw_value: bytes,
+        return_value: Any,
+        decode_error: Optional[Exception],
+        tx_info: dict,
+        method: abi.Method,
+    ) -> None:
+        self.tx_id = tx_id
+        self.raw_value = raw_value
+        self.return_value = return_value
+        self.decode_error = decode_error
+        self.tx_info = tx_info
+        self.method = method
+
+
+class AtomicTransactionResponse:
+    def __init__(
+        self, confirmed_round: int, tx_ids: List[str], results: List[ABIResult]
+    ) -> None:
+        self.confirmed_round = confirmed_round
+        self.tx_ids = tx_ids
+        self.abi_results = results
+
+
+class SimulateABIResult(ABIResult):
+    def __init__(
+        self,
+        tx_id: str,
+        raw_value: bytes,
+        return_value: Any,
+        decode_error: Optional[Exception],
+        tx_info: dict,
+        method: abi.Method,
+        missing_signature: bool,
+    ) -> None:
+        self.tx_id = tx_id
+        self.raw_value = raw_value
+        self.return_value = return_value
+        self.decode_error = decode_error
+        self.tx_info = tx_info
+        self.method = method
+        self.missing_signature = missing_signature
+
+
+class SimulateAtomicTransactionResponse:
+    def __init__(
+        self,
+        version: int,
+        would_succeed: bool,
+        failure_message: str,
+        failed_at: Optional[List[int]],
+        simulate_response: Dict[str, Any],
+        tx_ids: List[str],
+        results: List[SimulateABIResult],
+    ) -> None:
+        self.version = version
+        self.would_succeed = would_succeed
+        self.failure_message = failure_message
+        self.failed_at = failed_at
+        self.simulate_response = simulate_response
+        self.tx_ids = tx_ids
+        self.abi_results = results
 
 
 class AtomicTransactionComposer:
@@ -102,7 +305,9 @@ class AtomicTransactionComposer:
     MAX_APP_ARG_LIMIT = 16
 
     def __init__(self) -> None:
-        self.status = AtomicTransactionComposerStatus.BUILDING
+        self.status: AtomicTransactionComposerStatus = (
+            AtomicTransactionComposerStatus.BUILDING
+        )
         self.method_dict: Dict[int, abi.Method] = {}
         self.txn_list: List[TransactionWithSigner] = []
         self.signed_txns: List[GenericSignedTransaction] = []
@@ -135,7 +340,7 @@ class AtomicTransactionComposer:
         return cloned
 
     def add_transaction(
-        self, txn_and_signer: "TransactionWithSigner"
+        self, txn_and_signer: TransactionWithSigner
     ) -> "AtomicTransactionComposer":
         """
         Adds a transaction to this atomic group.
@@ -174,16 +379,14 @@ class AtomicTransactionComposer:
         method: abi.Method,
         sender: str,
         sp: transaction.SuggestedParams,
-        signer: "TransactionSigner",
-        method_args: Optional[
-            List[Union[Any, "TransactionWithSigner"]]
-        ] = None,
+        signer: TransactionSigner,
+        method_args: Optional[List[Union[Any, TransactionWithSigner]]] = None,
         on_complete: transaction.OnComplete = transaction.OnComplete.NoOpOC,
         local_schema: Optional[transaction.StateSchema] = None,
         global_schema: Optional[transaction.StateSchema] = None,
         approval_program: Optional[bytes] = None,
         clear_program: Optional[bytes] = None,
-        extra_pages: Optional[int] = None,
+        extra_pages: int = 0,
         accounts: Optional[List[str]] = None,
         foreign_apps: Optional[List[int]] = None,
         foreign_assets: Optional[List[int]] = None,
@@ -381,7 +584,7 @@ class AtomicTransactionComposer:
         self.method_dict[len(self.txn_list) - 1] = method
         return self
 
-    def build_group(self) -> list:
+    def build_group(self) -> List[TransactionWithSigner]:
         """
         Finalize the transaction group and returns the finalized transactions with signers.
         The composer's status will be at least BUILT after executing this method.
@@ -409,7 +612,7 @@ class AtomicTransactionComposer:
         self.status = AtomicTransactionComposerStatus.BUILT
         return self.txn_list
 
-    def gather_signatures(self) -> list:
+    def gather_signatures(self) -> List[GenericSignedTransaction]:
         """
         Obtain signatures for each transaction in this group. If signatures have already been obtained,
         this method will return cached versions of the signatures.
@@ -479,9 +682,92 @@ class AtomicTransactionComposer:
         self.status = AtomicTransactionComposerStatus.SUBMITTED
         return self.tx_ids
 
+    def simulate(
+        self, client: algod.AlgodClient
+    ) -> SimulateAtomicTransactionResponse:
+        """
+        Send the transaction group to the `simulate` endpoint and wait for results.
+        An error will be thrown if submission or execution fails.
+        The composer's status must be SUBMITTED or lower before calling this method,
+        since execution is only allowed once.
+
+        Args:
+            client (AlgodClient): Algod V2 client
+
+        Returns:
+            SimulateAtomicTransactionResponse: Object with simulation results for this
+                transaction group, a list of txIDs of the simulated transactions,
+                an array of results for each method call transaction in this group.
+                If a method has no return value (void), then the method results array
+                will contain None for that method's return value.
+        """
+
+        if self.status <= AtomicTransactionComposerStatus.SUBMITTED:
+            self.gather_signatures()
+        else:
+            raise error.AtomicTransactionComposerError(
+                "AtomicTransactionComposerStatus must be submitted or "
+                "lower to simulate a group"
+            )
+
+        simulation_result = cast(
+            Dict[str, Any], client.simulate_transactions(self.signed_txns)
+        )
+        # Only take the first group in the simulate response
+        txn_group: Dict[str, Any] = simulation_result["txn-groups"][0]
+
+        # Parse out abi results
+        txn_results = [t["txn-result"] for t in txn_group["txn-results"]]
+        method_results: List[ABIResult] = []
+        for method_index, method in self.method_dict.items():
+            tx_id = self.tx_ids[method_index]
+            tx_info = txn_results[method_index]
+            result: ABIResult = ABIResult(
+                tx_id=tx_id,
+                raw_value=bytes(),
+                return_value=None,
+                decode_error=None,
+                tx_info=tx_info,
+                method=method,
+            )
+            try:
+                result = self.parse_result(
+                    method, self.tx_ids[method_index], tx_info
+                )
+            except Exception as e:
+                result.decode_error = e
+            method_results.append(result)
+
+        # build up data structure with fields we'd want
+        sim_results = []
+        for idx, result in enumerate(method_results):
+            sim_txn: Dict[str, Any] = txn_group["txn-results"][idx]
+
+            sim_results.append(
+                SimulateABIResult(
+                    tx_id=result.tx_id,
+                    raw_value=result.raw_value,
+                    return_value=result.return_value,
+                    decode_error=result.decode_error,
+                    tx_info=result.tx_info,
+                    method=result.method,
+                    missing_signature=sim_txn.get("missing-signature", False),
+                )
+            )
+
+        return SimulateAtomicTransactionResponse(
+            version=simulation_result.get("version", 0),
+            would_succeed=simulation_result.get("would-succeed", False),
+            failure_message=txn_group.get("failure-message", ""),
+            failed_at=txn_group.get("failed-at"),
+            simulate_response=simulation_result,
+            tx_ids=self.tx_ids,
+            results=sim_results,
+        )
+
     def execute(
         self, client: algod.AlgodClient, wait_rounds: int
-    ) -> "AtomicTransactionResponse":
+    ) -> AtomicTransactionResponse:
         """
         Send the transaction group to the network and wait until it's committed
         to a block. An error will be thrown if submission or execution fails.
@@ -517,67 +803,28 @@ class AtomicTransactionComposer:
         self.status = AtomicTransactionComposerStatus.COMMITTED
 
         confirmed_round = resp["confirmed-round"]
-        method_results = []
+        method_results: List[ABIResult] = []
 
-        for i, tx_id in enumerate(self.tx_ids):
-            raw_value: Optional[bytes] = None
-            return_value = None
-            decode_error = None
-            tx_info: Optional[Any] = None
-
-            if i not in self.method_dict:
-                continue
-
-            # Parse log for ABI method return value
-            try:
-                tx_info = client.pending_transaction_info(tx_id)
-                if self.method_dict[i].returns.type == abi.Returns.VOID:
-                    method_results.append(
-                        ABIResult(
-                            tx_id=tx_id,
-                            raw_value=cast(bytes, raw_value),
-                            return_value=return_value,
-                            decode_error=decode_error,
-                            tx_info=tx_info,
-                            method=self.method_dict[i],
-                        )
-                    )
-                    continue
-
-                logs = tx_info["logs"] if "logs" in tx_info else []
-
-                # Look for the last returned value in the log
-                if not logs:
-                    raise error.AtomicTransactionComposerError(
-                        "app call transaction did not log a return value"
-                    )
-                result = logs[-1]
-                # Check that the first four bytes is the hash of "return"
-                result_bytes = base64.b64decode(result)
-                if (
-                    len(result_bytes) < 4
-                    or result_bytes[:4] != ABI_RETURN_HASH
-                ):
-                    raise error.AtomicTransactionComposerError(
-                        "app call transaction did not log a return value"
-                    )
-                raw_value = result_bytes[4:]
-                method_return_type = cast(
-                    abi.ABIType, self.method_dict[i].returns.type
-                )
-                return_value = method_return_type.decode(raw_value)
-            except Exception as e:
-                decode_error = e
-
-            abi_result = ABIResult(
+        for method_index, method in self.method_dict.items():
+            tx_id = self.tx_ids[method_index]
+            result: ABIResult = ABIResult(
                 tx_id=tx_id,
-                raw_value=cast(bytes, raw_value),
-                return_value=return_value,
-                decode_error=decode_error,
-                tx_info=cast(Any, tx_info),
-                method=self.method_dict[i],
+                raw_value=bytes(),
+                return_value=None,
+                decode_error=None,
+                tx_info={},
+                method=method,
             )
-            method_results.append(abi_result)
+            try:
+                tx_info = cast(
+                    Dict[str, Any], client.pending_transaction_info(tx_id)
+                )
+                result = self.parse_result(
+                    method, self.tx_ids[method_index], tx_info
+                )
+            except Exception as e:
+                result.decode_error = e
+            method_results.append(result)
 
         return AtomicTransactionResponse(
             confirmed_round=confirmed_round,
@@ -585,158 +832,49 @@ class AtomicTransactionComposer:
             results=method_results,
         )
 
+    def parse_result(
+        self, method: abi.Method, txid: str, txn: Dict[str, Any]
+    ) -> ABIResult:
+        tx_id = txid
+        raw_value = bytes()
+        return_value = None
+        decode_error = None
+        try:
+            if method.returns.type == abi.Returns.VOID:
+                return ABIResult(
+                    tx_id=tx_id,
+                    raw_value=raw_value,
+                    return_value=return_value,
+                    decode_error=decode_error,
+                    tx_info=txn,
+                    method=method,
+                )
 
-class TransactionSigner(ABC):
-    """
-    Represents an object which can sign transactions from an atomic transaction group.
-    """
+            logs = txn["logs"] if "logs" in txn else []
 
-    def __init__(self) -> None:
-        pass
+            # Look for the last returned value in the log
+            if not logs:
+                raise error.AtomicTransactionComposerError(
+                    "app call transaction did not log a return value"
+                )
+            result = logs[-1]
+            # Check that the first four bytes is the hash of "return"
+            result_bytes = base64.b64decode(result)
+            if len(result_bytes) < 4 or result_bytes[:4] != ABI_RETURN_HASH:
+                raise error.AtomicTransactionComposerError(
+                    "app call transaction did not log a return value"
+                )
+            raw_value = result_bytes[4:]
+            method_return_type = cast(abi.ABIType, method.returns.type)
+            return_value = method_return_type.decode(raw_value)
+        except Exception as e:
+            decode_error = e
 
-    @abstractmethod
-    def sign_transactions(
-        self, txn_group: List[transaction.Transaction], indexes: List[int]
-    ) -> List[GenericSignedTransaction]:
-        pass
-
-
-class AccountTransactionSigner(TransactionSigner):
-    """
-    Represents a Transaction Signer for an account that can sign transactions from an
-    atomic transaction group.
-
-    Args:
-        private_key (str): private key of signing account
-    """
-
-    def __init__(self, private_key: str) -> None:
-        super().__init__()
-        self.private_key = private_key
-
-    def sign_transactions(
-        self, txn_group: List[transaction.Transaction], indexes: List[int]
-    ) -> List[GenericSignedTransaction]:
-        """
-        Sign transactions in a transaction group given the indexes.
-
-        Returns an array of encoded signed transactions. The length of the
-        array will be the same as the length of indexesToSign, and each index i in the array
-        corresponds to the signed transaction from txnGroup[indexesToSign[i]].
-
-        Args:
-            txn_group (list[Transaction]): atomic group of transactions
-            indexes (list[int]): array of indexes in the atomic transaction group that should be signed
-        """
-        stxns = []
-        for i in indexes:
-            stxn = txn_group[i].sign(self.private_key)
-            stxns.append(stxn)
-        return stxns
-
-
-class LogicSigTransactionSigner(TransactionSigner):
-    """
-    Represents a Transaction Signer for a LogicSig that can sign transactions from an
-    atomic transaction group.
-
-    Args:
-        lsig (LogicSigAccount): LogicSig account
-    """
-
-    def __init__(self, lsig: transaction.LogicSigAccount) -> None:
-        super().__init__()
-        self.lsig = lsig
-
-    def sign_transactions(
-        self, txn_group: List[transaction.Transaction], indexes: List[int]
-    ) -> List[GenericSignedTransaction]:
-        """
-        Sign transactions in a transaction group given the indexes.
-
-        Returns an array of encoded signed transactions. The length of the
-        array will be the same as the length of indexesToSign, and each index i in the array
-        corresponds to the signed transaction from txnGroup[indexesToSign[i]].
-
-        Args:
-            txn_group (list[Transaction]): atomic group of transactions
-            indexes (list[int]): array of indexes in the atomic transaction group that should be signed
-        """
-        stxns: List[GenericSignedTransaction] = []
-        for i in indexes:
-            stxn = transaction.LogicSigTransaction(txn_group[i], self.lsig)
-            stxns.append(stxn)
-        return stxns
-
-
-class MultisigTransactionSigner(TransactionSigner):
-    """
-    Represents a Transaction Signer for a Multisig that can sign transactions from an
-    atomic transaction group.
-
-    Args:
-        msig (Multisig): Multisig account
-        sks (str): private keys of multisig
-    """
-
-    def __init__(self, msig: transaction.Multisig, sks: str) -> None:
-        super().__init__()
-        self.msig = msig
-        self.sks = sks
-
-    def sign_transactions(
-        self, txn_group: List[transaction.Transaction], indexes: List[int]
-    ) -> List[GenericSignedTransaction]:
-        """
-        Sign transactions in a transaction group given the indexes.
-
-        Returns an array of encoded signed transactions. The length of the
-        array will be the same as the length of indexesToSign, and each index i in the array
-        corresponds to the signed transaction from txnGroup[indexesToSign[i]].
-
-        Args:
-            txn_group (list[Transaction]): atomic group of transactions
-            indexes (list[int]): array of indexes in the atomic transaction group that should be signed
-        """
-        stxns: List[GenericSignedTransaction] = []
-        for i in indexes:
-            mtxn = transaction.MultisigTransaction(txn_group[i], self.msig)
-            for sk in self.sks:
-                mtxn.sign(sk)
-            stxns.append(mtxn)
-        return stxns
-
-
-class TransactionWithSigner:
-    def __init__(
-        self, txn: transaction.Transaction, signer: TransactionSigner
-    ) -> None:
-        self.txn = txn
-        self.signer = signer
-
-
-class ABIResult:
-    def __init__(
-        self,
-        tx_id: str,
-        raw_value: bytes,
-        return_value: Any,
-        decode_error: Optional[Exception],
-        tx_info: dict,
-        method: abi.Method,
-    ) -> None:
-        self.tx_id = tx_id
-        self.raw_value = raw_value
-        self.return_value = return_value
-        self.decode_error = decode_error
-        self.tx_info = tx_info
-        self.method = method
-
-
-class AtomicTransactionResponse:
-    def __init__(
-        self, confirmed_round: int, tx_ids: List[str], results: List[ABIResult]
-    ) -> None:
-        self.confirmed_round = confirmed_round
-        self.tx_ids = tx_ids
-        self.abi_results = results
+        return ABIResult(
+            tx_id=tx_id,
+            raw_value=raw_value,
+            return_value=return_value,
+            decode_error=decode_error,
+            tx_info=txn,
+            method=method,
+        )
