@@ -1,11 +1,29 @@
 import base64
 import json
+from typing import (
+    Any,
+    Dict,
+    Final,
+    Iterable,
+    List,
+    Mapping,
+    Optional,
+    Sequence,
+    Tuple,
+    Union,
+    cast,
+)
 import urllib.error
 from urllib import parse
 from urllib.request import Request, urlopen
 
 from algosdk import constants, encoding, error, transaction, util
 
+AlgodResponseType = Union[Dict[str, Any], bytes]
+
+# for compatibility with urllib.parse.urlencode
+ParamsType = Union[Mapping[str, Any], Sequence[Tuple[str, Any]]]
+
 api_version_path_prefix = "/v2"
 
 
@@ -24,32 +42,39 @@ class AlgodClient:
         headers (dict)
     """
 
-    def __init__(self, algod_token, algod_address, headers=None):
-        self.algod_token = algod_token
-        self.algod_address = algod_address
-        self.headers = headers
+    def __init__(
+        self,
+        algod_token: str,
+        algod_address: str,
+        headers: Optional[Dict[str, str]] = None,
+    ):
+        self.algod_token: Final[str] = algod_token
+        self.algod_address: Final[str] = algod_address
+        self.headers: Final[Optional[Dict[str, str]]] = headers
 
     def algod_request(
         self,
-        method,
-        requrl,
-        params=None,
-        data=None,
-        headers=None,
-        response_format="json",
-    ):
+        method: str,
+        requrl: str,
+        params: Optional[ParamsType] = None,
+        data: Optional[bytes] = None,
+        headers: Optional[Dict[str, str]] = None,
+        response_format: Optional[str] = "json",
+    ) -> AlgodResponseType:
         """
         Execute a given request.
 
         Args:
             method (str): request method
             requrl (str): url for the request
-            params (dict, optional): parameters for the request
-            data (dict, optional): data in the body of the request
+            params (ParamsType, optional): parameters for the request
+            data (bytes, optional): data in the body of the request
             headers (dict, optional): additional header for request
+            response_format (str, optional): format of the response
 
         Returns:
-            dict: loaded from json response body
+            dict loaded from json response body when response_format == "json"
+            otherwise returns the response body as bytes
         """
         header = {"User-Agent": "py-algorand-sdk"}
 
@@ -78,9 +103,9 @@ class AlgodClient:
             resp = urlopen(req)
         except urllib.error.HTTPError as e:
             code = e.code
-            e = e.read().decode("utf-8")
+            es = e.read().decode("utf-8")
             try:
-                e = json.loads(e)["message"]
+                e = json.loads(es)["message"]
             finally:
                 raise error.AlgodHTTPError(e, code)
         if response_format == "json":
@@ -93,7 +118,18 @@ class AlgodClient:
         else:
             return resp.read()
 
-    def account_info(self, address, exclude=None, **kwargs):
+    @classmethod
+    def _assert_json_response(
+        cls, params: Mapping[str, Any], endpoint: str = ""
+    ) -> None:
+        if params.get("response_format", "json") != "json":
+            raise error.AlgodRequestError(
+                f"Only json response is supported{ (' for ' + endpoint) if endpoint else ''}."
+            )
+
+    def account_info(
+        self, address: str, exclude: Optional[str] = None, **kwargs: Any
+    ) -> AlgodResponseType:
         """
         Return account information.
 
@@ -106,7 +142,7 @@ class AlgodClient:
         req = "/accounts/" + address
         return self.algod_request("GET", req, query, **kwargs)
 
-    def asset_info(self, asset_id, **kwargs):
+    def asset_info(self, asset_id: int, **kwargs: Any) -> AlgodResponseType:
         """
         Return information about a specific asset.
 
@@ -116,7 +152,9 @@ class AlgodClient:
         req = "/assets/" + str(asset_id)
         return self.algod_request("GET", req, **kwargs)
 
-    def application_info(self, application_id, **kwargs):
+    def application_info(
+        self, application_id: int, **kwargs: Any
+    ) -> AlgodResponseType:
         """
         Return information about a specific application.
 
@@ -127,8 +165,8 @@ class AlgodClient:
         return self.algod_request("GET", req, **kwargs)
 
     def application_box_by_name(
-        self, application_id: int, box_name: bytes, **kwargs
-    ):
+        self, application_id: int, box_name: bytes, **kwargs: Any
+    ) -> AlgodResponseType:
         """
         Return the value of an application's box.
 
@@ -144,7 +182,9 @@ class AlgodClient:
         params = {"name": box_name_encoded}
         return self.algod_request("GET", req, params=params, **kwargs)
 
-    def application_boxes(self, application_id: int, limit: int = 0, **kwargs):
+    def application_boxes(
+        self, application_id: int, limit: int = 0, **kwargs: Any
+    ) -> AlgodResponseType:
         """
         Given an application ID, return all Box names. No particular ordering is guaranteed. Request fails when client or server-side configured limits prevent returning all Box names.
 
@@ -159,7 +199,9 @@ class AlgodClient:
         params = {"max": limit} if limit else {}
         return self.algod_request("GET", req, params=params, **kwargs)
 
-    def account_asset_info(self, address, asset_id, **kwargs):
+    def account_asset_info(
+        self, address: str, asset_id: int, **kwargs: Any
+    ) -> AlgodResponseType:
         """
         Return asset information for a specific account.
 
@@ -167,11 +209,13 @@ class AlgodClient:
             address (str): account public key
             asset_id (int): The ID of the asset to look up.
         """
-        query = {}
+        query: Mapping = {}
         req = "/accounts/" + address + "/assets/" + str(asset_id)
         return self.algod_request("GET", req, query, **kwargs)
 
-    def account_application_info(self, address, application_id, **kwargs):
+    def account_application_info(
+        self, address: str, application_id: int, **kwargs: Any
+    ) -> AlgodResponseType:
         """
         Return application information for a specific account.
 
@@ -179,13 +223,17 @@ class AlgodClient:
             address (str): account public key
             application_id (int): The ID of the application to look up.
         """
-        query = {}
+        query: Mapping = {}
         req = "/accounts/" + address + "/applications/" + str(application_id)
         return self.algod_request("GET", req, query, **kwargs)
 
     def pending_transactions_by_address(
-        self, address, limit=0, response_format="json", **kwargs
-    ):
+        self,
+        address: str,
+        limit: int = 0,
+        response_format: str = "json",
+        **kwargs: Any,
+    ) -> AlgodResponseType:
         """
         Get the list of pending transactions by address, sorted by priority,
         in decreasing order, truncated at the end at MAX. If MAX = 0, returns
@@ -197,7 +245,7 @@ class AlgodClient:
             response_format (str): the format in which the response is returned: either
                 "json" or "msgpack"
         """
-        query = {"format": response_format}
+        query: Dict[str, Union[str, int]] = {"format": response_format}
         if limit:
             query["max"] = limit
         req = "/accounts/" + address + "/transactions/pending"
@@ -207,8 +255,12 @@ class AlgodClient:
         return res
 
     def block_info(
-        self, block=None, response_format="json", round_num=None, **kwargs
-    ):
+        self,
+        block: Optional[int] = None,
+        response_format: str = "json",
+        round_num: Optional[int] = None,
+        **kwargs: Any,
+    ) -> AlgodResponseType:
         """
         Get the block for the given round.
 
@@ -219,25 +271,28 @@ class AlgodClient:
             round_num (int, optional): alias for block; specify one of these
         """
         query = {"format": response_format}
-        if block is None and round_num is None:
-            raise error.UnderspecifiedRoundError
         req = "/blocks/" + _specify_round_string(block, round_num)
         res = self.algod_request(
             "GET", req, query, response_format=response_format, **kwargs
         )
         return res
 
-    def ledger_supply(self, **kwargs):
+    def ledger_supply(self, **kwargs: Any) -> AlgodResponseType:
         """Return supply details for node's ledger."""
         req = "/ledger/supply"
         return self.algod_request("GET", req, **kwargs)
 
-    def status(self, **kwargs):
+    def status(self, **kwargs: Any) -> AlgodResponseType:
         """Return node status."""
         req = "/status"
         return self.algod_request("GET", req, **kwargs)
 
-    def status_after_block(self, block_num=None, round_num=None, **kwargs):
+    def status_after_block(
+        self,
+        block_num: Optional[int] = None,
+        round_num: Optional[int] = None,
+        **kwargs: Any,
+    ) -> AlgodResponseType:
         """
         Return node status immediately after blockNum.
 
@@ -246,14 +301,14 @@ class AlgodClient:
             round_num (int, optional): alias for block_num; specify one of
                 these
         """
-        if block_num is None and round_num is None:
-            raise error.UnderspecifiedRoundError
         req = "/status/wait-for-block-after/" + _specify_round_string(
             block_num, round_num
         )
         return self.algod_request("GET", req, **kwargs)
 
-    def send_transaction(self, txn, **kwargs):
+    def send_transaction(
+        self, txn: "transaction.Transaction", **kwargs: Any
+    ) -> str:
         """
         Broadcast a signed transaction object to the network.
 
@@ -266,12 +321,14 @@ class AlgodClient:
         """
         assert not isinstance(
             txn, transaction.Transaction
-        ), "Attempt to send UNSIGNED transaction {}".format(txn)
+        ), "Attempt to send UNSUPPORTED type of transaction {}".format(txn)
         return self.send_raw_transaction(
             encoding.msgpack_encode(txn), **kwargs
         )
 
-    def send_raw_transaction(self, txn, **kwargs):
+    def send_raw_transaction(
+        self, txn: Union[bytes, str], **kwargs: Any
+    ) -> str:
         """
         Broadcast a signed transaction to the network.
 
@@ -282,7 +339,9 @@ class AlgodClient:
         Returns:
             str: transaction ID
         """
-        txn = base64.b64decode(txn)
+        self._assert_json_response(kwargs, "send_raw_transaction")
+
+        txn_bytes = base64.b64decode(txn)
         req = "/transactions"
         headers = util.build_headers_from(
             kwargs.get("headers", False),
@@ -290,11 +349,12 @@ class AlgodClient:
         )
         kwargs["headers"] = headers
 
-        return self.algod_request("POST", req, data=txn, **kwargs)["txId"]
+        resp = self.algod_request("POST", req, data=txn_bytes, **kwargs)
+        return cast(str, cast(dict, resp)["txId"])
 
     def pending_transactions(
-        self, max_txns=0, response_format="json", **kwargs
-    ):
+        self, max_txns: int = 0, response_format: str = "json", **kwargs: Any
+    ) -> AlgodResponseType:
         """
         Return pending transactions.
 
@@ -304,18 +364,17 @@ class AlgodClient:
             response_format (str): the format in which the response is returned: either
                 "json" or "msgpack"
         """
-        query = {"format": response_format}
+        query: Dict[str, Union[int, str]] = {"format": response_format}
         if max_txns:
             query["max"] = max_txns
         req = "/transactions/pending"
-        res = self.algod_request(
+        return self.algod_request(
             "GET", req, params=query, response_format=response_format, **kwargs
         )
-        return res
 
     def pending_transaction_info(
-        self, transaction_id, response_format="json", **kwargs
-    ):
+        self, transaction_id: str, response_format: str = "json", **kwargs: Any
+    ) -> AlgodResponseType:
         """
         Return transaction information for a pending transaction.
 
@@ -326,22 +385,25 @@ class AlgodClient:
         """
         req = "/transactions/pending/" + transaction_id
         query = {"format": response_format}
-        res = self.algod_request(
+        return self.algod_request(
             "GET", req, params=query, response_format=response_format, **kwargs
         )
-        return res
 
-    def health(self, **kwargs):
+    def health(self, **kwargs: Any) -> AlgodResponseType:
         """Return null if the node is running."""
         req = "/health"
         return self.algod_request("GET", req, **kwargs)
 
-    def versions(self, **kwargs):
+    def versions(self, **kwargs: Any) -> AlgodResponseType:
         """Return algod versions."""
         req = "/versions"
         return self.algod_request("GET", req, **kwargs)
 
-    def send_transactions(self, txns, **kwargs):
+    def send_transactions(
+        self,
+        txns: "Iterable[transaction.GenericSignedTransaction]",
+        **kwargs: Any,
+    ) -> str:
         """
         Broadcast list of a signed transaction objects to the network.
 
@@ -353,21 +415,22 @@ class AlgodClient:
         Returns:
             str: first transaction ID
         """
-        serialized = []
+        serialized: List[bytes] = []
         for txn in txns:
             assert not isinstance(
                 txn, transaction.Transaction
             ), "Attempt to send UNSIGNED transaction {}".format(txn)
             serialized.append(base64.b64decode(encoding.msgpack_encode(txn)))
-
         return self.send_raw_transaction(
             base64.b64encode(b"".join(serialized)), **kwargs
         )
 
-    def suggested_params(self, **kwargs):
+    def suggested_params(self, **kwargs: Any) -> "transaction.SuggestedParams":
         """Return suggested transaction parameters."""
+        self._assert_json_response(kwargs, "suggested_params")
+
         req = "/transactions/params"
-        res = self.algod_request("GET", req, **kwargs)
+        res = cast(dict, self.algod_request("GET", req, **kwargs))
 
         return transaction.SuggestedParams(
             res["fee"],
@@ -380,7 +443,9 @@ class AlgodClient:
             res["min-fee"],
         )
 
-    def compile(self, source, source_map=False, **kwargs):
+    def compile(
+        self, source: str, source_map: bool = False, **kwargs: Any
+    ) -> Dict[str, Any]:
         """
         Compile TEAL source with remote algod.
 
@@ -390,8 +455,9 @@ class AlgodClient:
 
         Returns:
             dict: loaded from json response body. "result" property contains compiled bytes, "hash" - program hash (escrow address)
-
         """
+        self._assert_json_response(kwargs, "compile")
+
         req = "/teal/compile"
         headers = util.build_headers_from(
             kwargs.get("headers", False),
@@ -399,11 +465,47 @@ class AlgodClient:
         )
         kwargs["headers"] = headers
         params = {"sourcemap": source_map}
-        return self.algod_request(
-            "POST", req, params=params, data=source.encode("utf-8"), **kwargs
+        return cast(
+            Dict[str, Any],
+            self.algod_request(
+                "POST",
+                req,
+                params=params,
+                data=source.encode("utf-8"),
+                **kwargs,
+            ),
         )
 
-    def dryrun(self, drr, **kwargs):
+    def disassemble(
+        self, program_bytes: bytes, **kwargs: Any
+    ) -> Dict[str, str]:
+        """
+        Disassable TEAL program bytes with remote algod.
+        Args:
+            program (bytes): bytecode to be disassembled
+            request_header (dict, optional): additional header for request
+        Returns:
+            dict: response dictionary containing disassembled TEAL source code
+            in plain text as the value of the unique "result" key.
+        """
+        if not isinstance(program_bytes, bytes):
+            raise error.InvalidProgram(
+                message=f"disassemble endpoints only accepts bytes but request program_bytes is of type {type(program_bytes)}"
+            )
+        self._assert_json_response(kwargs, "disassemble")
+
+        req = "/teal/disassemble"
+        headers = util.build_headers_from(
+            kwargs.get("headers", False),
+            {"Content-Type": "application/x-binary"},
+        )
+        kwargs["headers"] = headers
+        return cast(
+            Dict[str, str],
+            self.algod_request("POST", req, data=program_bytes, **kwargs),
+        )
+
+    def dryrun(self, drr: Dict[str, Any], **kwargs: Any) -> Dict[str, Any]:
         """
         Dryrun with remote algod.
 
@@ -414,6 +516,8 @@ class AlgodClient:
         Returns:
             dict: loaded from json response body
         """
+        self._assert_json_response(kwargs, "dryrun")
+
         req = "/teal/dryrun"
         headers = util.build_headers_from(
             kwargs.get("headers", False),
@@ -423,16 +527,21 @@ class AlgodClient:
         data = encoding.msgpack_encode(drr)
         data = base64.b64decode(data)
 
-        return self.algod_request("POST", req, data=data, **kwargs)
+        return cast(dict, self.algod_request("POST", req, data=data, **kwargs))
 
-    def genesis(self, **kwargs):
+    def genesis(self, **kwargs: Any) -> AlgodResponseType:
         """Returns the entire genesis file."""
         req = "/genesis"
         return self.algod_request("GET", req, **kwargs)
 
     def transaction_proof(
-        self, round_num, txid, hashtype="", response_format="json", **kwargs
-    ):
+        self,
+        round_num: int,
+        txid: str,
+        hashtype: str = "",
+        response_format: str = "json",
+        **kwargs: Any,
+    ) -> AlgodResponseType:
         """
         Get a proof for a transaction in a block.
 
@@ -450,12 +559,14 @@ class AlgodClient:
             req,
             params=params,
             response_format=response_format,
-            **kwargs
+            **kwargs,
         )
 
-    def lightblockheader_proof(self, round_num, **kwargs):
+    def lightblockheader_proof(
+        self, round_num: int, **kwargs: Any
+    ) -> AlgodResponseType:
         """
-         Gets a proof for a given light block header inside a state proof commitment.
+        Gets a proof for a given light block header inside a state proof commitment.
 
         Args:
             round_num (int): The round to which the light block header belongs.
@@ -463,7 +574,7 @@ class AlgodClient:
         req = "/blocks/{}/lightheader/proof".format(round_num)
         return self.algod_request("GET", req, **kwargs)
 
-    def stateproofs(self, round_num, **kwargs):
+    def stateproofs(self, round_num: int, **kwargs: Any) -> AlgodResponseType:
         """
         Get a state proof that covers a given round
 
@@ -473,7 +584,9 @@ class AlgodClient:
         req = "/stateproofs/{}".format(round_num)
         return self.algod_request("GET", req, **kwargs)
 
-    def get_block_hash(self, round_num, **kwargs):
+    def get_block_hash(
+        self, round_num: int, **kwargs: Any
+    ) -> AlgodResponseType:
         """
         Get the block hash for the block on the given round.
 
@@ -483,8 +596,55 @@ class AlgodClient:
         req = "/blocks/{}/hash".format(round_num)
         return self.algod_request("GET", req, **kwargs)
 
+    def simulate_transactions(
+        self,
+        txns: "Iterable[transaction.GenericSignedTransaction]",
+        **kwargs: Any,
+    ) -> AlgodResponseType:
+        """
+        Simulate a list of a signed transaction objects being sent to the network.
 
-def _specify_round_string(block, round_num):
+        Args:
+            txns (SignedTransaction[] or MultisigTransaction[]):
+                transactions to send
+            request_header (dict, optional): additional header for request
+
+        Returns:
+            Dict[str, Any]: results from simulation of transaction group
+        """
+        serialized = []
+        for txn in txns:
+            serialized.append(base64.b64decode(encoding.msgpack_encode(txn)))
+
+        return self.simulate_raw_transaction(
+            base64.b64encode(b"".join(serialized)), **kwargs
+        )
+
+    def simulate_raw_transaction(self, txn, **kwargs):
+        """
+        Simulate a transaction group
+
+        Args:
+            txn (str): transaction to send, encoded in base64
+            request_header (dict, optional): additional header for request
+
+        Returns:
+            Dict[str, Any]: results from simulation of transaction group
+        """
+        txn = base64.b64decode(txn)
+        req = "/transactions/simulate"
+        headers = util.build_headers_from(
+            kwargs.get("headers", False),
+            {"Content-Type": "application/x-binary"},
+        )
+        kwargs["headers"] = headers
+
+        return self.algod_request("POST", req, data=txn, **kwargs)
+
+
+def _specify_round_string(
+    block: Union[int, None], round_num: Union[int, None]
+) -> str:
     """
     Return the round number specified in either 'block' or 'round_num'.
 
@@ -492,10 +652,13 @@ def _specify_round_string(block, round_num):
         block (int): user specified variable
         round_num (int): user specified variable
     """
+    if block is None and round_num is None:
+        raise error.UnderspecifiedRoundError()
 
     if block is not None and round_num is not None:
-        raise error.OverspecifiedRoundError
-    elif block is not None:
-        return str(block)
-    elif round_num is not None:
+        raise error.OverspecifiedRoundError()
+
+    if round_num is not None:
         return str(round_num)
+
+    return str(block)
